;; Generated package description from mastodon.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "mastodon" "2.1.0.0.20260808.72" "Client for fediverse services using the Mastodon API" '((emacs "29.1") (persist "0.8") (tp "0.8")) :commit "5ddf28b9454ad9ee12738f9467e3ad013b788ee3" :authors '(("Johnson Denen" . "johnson.denen@gmail.com") ("Marty Hiatt" . "martianh@disroot.org")) :maintainer '("Marty Hiatt" . "martianh@disroot.org") :url "https://codeberg.org/martianh/mastodon.el")
